/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "draw.pmh"

#define	FE_DCH_CACHE_DEBUG	FALSE

namespace fe
{

DrawCached::DrawCached(void):
	m_inputQueue(0),
	m_outputQueue(0),
	m_queues(2)
{
	m_pCommandQueue=new CommandQueue[m_queues];

	m_grain[Command::e_null]=1;
	m_grain[Command::e_points]=1;
	m_grain[Command::e_lines]=2;
	m_grain[Command::e_triangles]=3;
	m_grain[Command::e_rectangles]=4;
	m_grain[Command::e_sphere]=1;
	m_grain[Command::e_cylinder]=1;
	m_grain[Command::e_cylinders]=1;
	m_grain[Command::e_raster]=1;
	m_grain[Command::e_text]=1;

	m_oversize[Command::e_null]=1;
	m_oversize[Command::e_points]=1;
	m_oversize[Command::e_lines]=1;
	m_oversize[Command::e_triangles]=1;
	m_oversize[Command::e_rectangles]=1;
	m_oversize[Command::e_sphere]=1;
	m_oversize[Command::e_cylinder]=1;
#if FE_DOUBLE_REAL
	m_oversize[Command::e_cylinders]=4;
#else
	m_oversize[Command::e_cylinders]=
			sizeof(SpatialTransform)/sizeof(SpatialVector);
#endif
	m_oversize[Command::e_raster]=1;
	m_oversize[Command::e_text]=1;
}

DrawCached::~DrawCached(void)
{
	delete[] m_pCommandQueue;
}

void DrawCached::initialize(void)
{
}

void DrawCached::flush(void)
{
	clearCameraDisplayMap();

	m_outputQueue=m_inputQueue;
	m_inputQueue=(m_inputQueue+1)%m_queues;

	drawOutput(m_outputQueue,TRUE);
}

void DrawCached::clearInput(void)
{
	CommandQueue& rCommandQueue=m_pCommandQueue[m_inputQueue];
	U32& rCommands=rCommandQueue.m_commands;
	rCommands=0;

	clearCameraDisplayMap();
}

BWORD DrawCached::empty(void) const
{
	CommandQueue& rCommandQueue=m_pCommandQueue[m_inputQueue];
	return !rCommandQueue.m_commands;
}

void DrawCached::flushLive(void)
{
	drawOutput(m_inputQueue,FALSE);
}

BWORD DrawCached::prepareChain(void)
{
	if(!m_spDrawChain.isValid())
	{
		const String hintChain=
				registry()->master()->catalog()->catalogOrDefault<String>(
				"hint:DrawChain","");
		if(!hintChain.empty())
		{
			m_spDrawChain=registry()->create(hintChain);
		}
		if(!m_spDrawChain.isValid() && hintChain!="*.DrawOpenGL")
		{
			m_spDrawChain=registry()->create("*.DrawOpenGL");
		}
		if(!m_spDrawChain.isValid())
		{
			feLog("DrawCached::initialize failed to create immediate DrawI\n");
			return FALSE;
		}
	}

	m_spDrawChain->view()->setWindow(m_spWindow);
	return TRUE;
}

void DrawCached::drawOutput(U32 queue,BWORD a_clearQueue)
{
	FEASSERT(queue<m_queues);

	if(!prepareChain())
	{
		return;
	}

	//* HACK need to do clear and swap once per every window that got input
	sp<WindowI> spWindowI=view()->window();
	if(spWindowI.isValid())
	{
		spWindowI->makeCurrent();
		spWindowI->swapBuffers();
		spWindowI->clear();
	}
	else
	{
//		feX("DrawCached::drawOutput","view()->window() not valid");
//		return;
	}

	if(m_spDrawChain->brightness() != brightness())
	{
		m_spDrawChain->setBrightness(brightness());

		//* write back in case it was corrected
		setBrightness(m_spDrawChain->brightness());
	}

	if(m_spDrawChain->contrast() != contrast())
	{
		m_spDrawChain->setContrast(contrast());

		//* write back in case it was corrected
		setContrast(m_spDrawChain->contrast());
	}

	CommandQueue& rCommandQueue=m_pCommandQueue[queue];
	U32& rCommands=rCommandQueue.m_commands;

#if FE_DCH_CACHE_DEBUG
		feLog("DrawCached::drawOutput \"%s\" queue %d commands %d\n",
				name().c_str(),queue,rCommands);
#endif

	if(!rCommands)
	{
		return;
	}

	I32 currentLayer=INT_MAX;
	for(U32 command=0;command<rCommands;command++)
	{
		const Command& rCommand=rCommandQueue[command];
		const I32 layer=rCommand.layer();
		if(currentLayer>layer)
		{
			currentLayer=layer;
		}
	}

	I32 nextLayer=currentLayer;
	U32 completed=0;
	while(completed<rCommands)
	{
		for(U32 command=0;command<rCommands;command++)
		{
			const Command& rCommand=rCommandQueue[command];
			const I32 layer=rCommand.layer();

			if(layer>currentLayer &&
					(nextLayer==currentLayer || nextLayer>layer))
			{
				nextLayer=layer;
			}

			if(layer!=currentLayer)
			{
				continue;
			}

#if FE_DCH_CACHE_DEBUG
			feLog("[36mDrawCached::drawOutput queue %d command %d layer %d\n"
					" vertices %d strip %d multicolor %d\n"
					" vertex %p normal %p color %p%s\n"
					" texture %p image %p drawable %p[0m\n",
					queue,command,
					rCommand.layer(),
					rCommand.m_vertices,
					rCommand.m_stripMode,rCommand.m_multicolor,
					rCommand.m_pVertex,rCommand.m_pNormal,
					rCommand.m_pColor,
					(rCommand.m_pColor && !rCommand.m_multicolor)?
					(" "+print(*rCommand.m_pColor)).c_str(): "",
					rCommand.m_pTexture,
					rCommand.m_spImageI.raw(),rCommand.m_cpDrawableI.raw());

/*
			for(U32 m=0;m<rCommand.m_vertices;m++)
			{
				feLog("vert %d/%d %s norm %s\n",m,rCommand.m_vertices,
						rCommand.m_pVertex?
						c_print(rCommand.m_pVertex[m]): "NULL",
						rCommand.m_pNormal?
						c_print(rCommand.m_pNormal[m]): "NULL");
			}
*/
#endif

			const sp<CameraDisplay>& rspCameraDisplay=
					rCommand.m_spCameraDisplay;
			if(!completed ||
					m_spDrawChain->view()->cameraDisplay()!=rspCameraDisplay)
			{
#if FE_DCH_CACHE_DEBUG
				feLog("DrawCached::drawOutput changing CameraDisplay\n");
#endif
				m_spDrawChain->view()->setCameraDisplay(rspCameraDisplay);
				m_spDrawChain->view()->use(rCommand.m_projection);
			}
			else if(m_spDrawChain->view()->cameraDisplay()->projection()!=
					rCommand.m_projection)
			{
				m_spDrawChain->view()->use(rCommand.m_projection);
#if FE_DCH_CACHE_DEBUG
				feLog("DrawCached::drawOutput just changing projection\n");
#endif
			}

			//* TODO secondary sort on DrawMode (minimize context switching)

			const sp<DrawMode>& rspDrawMode=rCommand.m_spDrawMode;
			if(m_spDrawChain->drawMode()!=rspDrawMode)
			{
#if FE_DCH_CACHE_DEBUG
				feLog("DrawCached::drawOutput changing DrawMode\n");
#endif

				m_spDrawChain->setDrawMode(rspDrawMode);
			}

#if FE_DCH_CACHE_DEBUG
			feLog("DrawCached::drawOutput CameraDisplay:\n  %s\n",
					print(rspCameraDisplay).c_str());
			feLog(" %.6G %.6G %.6G %.6G\n",
					rspCameraDisplay->viewport()[0],
					rspCameraDisplay->viewport()[1],
					rspCameraDisplay->viewport().size()[0],
					rspCameraDisplay->viewport().size()[1]);
			feLog("DrawCached::drawOutput DrawMode:\n  %s\n",
					print(rspDrawMode).c_str());
#endif

			switch(rCommand.m_instruction)
			{
				case Command::e_points:
				{
#if FE_DCH_CACHE_DEBUG
					feLog("  points\n");
#endif
					m_spDrawChain->drawPoints(rCommand.m_pVertex,
							rCommand.m_pNormal,
							rCommand.m_vertices,
							rCommand.m_multicolor,
							rCommand.m_pColor,

							//* TODO transient buffers
//							rCommand.m_spDrawBuffer);
							sp<DrawBufferI>(NULL));
				}
				break;

				case Command::e_lines:
				{
#if FE_DCH_CACHE_DEBUG
					feLog("  lines\n");
#endif
					m_spDrawChain->drawLines(rCommand.m_pVertex,
							rCommand.m_pNormal,
							rCommand.m_vertices,
							rCommand.m_stripMode,
							rCommand.m_multicolor,
							rCommand.m_pColor,
							rCommand.m_pElement,
							rCommand.m_elementCount,

							//* TODO transient buffers
//							rCommand.m_spDrawBuffer);
							sp<DrawBufferI>(NULL));
				}
				break;

				case Command::e_triangles:
				{
#if FE_DCH_CACHE_DEBUG
					feLog("  triangles\n");
#endif
					m_spDrawChain->drawTriangles(rCommand.m_pVertex,
							rCommand.m_pNormal,
							rCommand.m_pTexture,
							rCommand.m_vertices,
							rCommand.m_stripMode,
							rCommand.m_multicolor,
							rCommand.m_pColor,
							rCommand.m_pHullPointMap,
							rCommand.m_pHullFacePoint,

							//* TODO transient buffers
//							rCommand.m_spDrawBuffer);
							sp<DrawBufferI>(NULL));
				}
				break;

				case Command::e_rectangles:
				{
#if FE_DCH_CACHE_DEBUG
					feLog("  rectangles\n");
#endif
					m_spDrawChain->drawRectangles(rCommand.m_pVertex,
							rCommand.m_vertices,
							rCommand.m_multicolor,
							rCommand.m_pColor);
				}
				break;

				case Command::e_sphere:
				{
#if FE_DCH_CACHE_DEBUG
					feLog("  sphere\n");
#endif
					SpatialTransform transform;
					set(transform,&rCommand.m_pVertex[0][0]);

					m_spDrawChain->drawSphere(transform,
							&rCommand.m_pVertex[6],
							*rCommand.m_pColor);
				}
				break;

				case Command::e_cylinder:
				{
#if FE_DCH_CACHE_DEBUG
					feLog("  cylinder\n");
#endif
					SpatialTransform transform;
					set(transform,&rCommand.m_pVertex[0][0]);

					m_spDrawChain->drawCylinder(transform,
							&rCommand.m_pVertex[6],rCommand.m_pVertex[7][0],
							*rCommand.m_pColor,U32(rCommand.m_pVertex[7][1]));
				}
				break;

				case Command::e_cylinders:
				{
#if FE_DCH_CACHE_DEBUG
					feLog("  cylinders\n");
#endif
					CylinderOverlay* pCylinderOverlay=
							(CylinderOverlay*)rCommand.m_pTexture;

					m_spDrawChain->drawCylinders(
							(SpatialTransform*)rCommand.m_pVertex,
							rCommand.m_pNormal,
							pCylinderOverlay->m_pBaseScale,
							pCylinderOverlay->m_pSlices,
							rCommand.m_vertices,
							rCommand.m_multicolor,
							rCommand.m_pColor);
				}
				break;

				case Command::e_raster:
				{
#if FE_DCH_CACHE_DEBUG
					feLog("  raster\n");
#endif
					m_spDrawChain->drawRaster(rCommand.m_spImageI,
							rCommand.m_pVertex[0]);
				}
				break;

				case Command::e_drawable:
				{
#if FE_DCH_CACHE_DEBUG
					feLog("  drawable\n");
#endif
					SpatialTransform transform;
					set(transform,&rCommand.m_pVertex[0][0]);
					m_spDrawChain->drawTransformed(
							transform,
							rCommand.m_cpDrawableI,
							rCommand.m_pColor,
							rCommand.m_spDrawBuffer);
				}
				break;

				case Command::e_text:
				{
					char* buffer=reinterpret_cast<char*>(rCommand.m_pTexture);

#if FE_DCH_CACHE_DEBUG
					feLog("  text \"%s\"\n",buffer);
#endif
					m_spDrawChain->drawAlignedText(rCommand.m_pVertex[0],
							buffer,rCommand.m_pColor[0]);
				}
				break;

				default:
				break;
			}

			completed++;
		}

		if(completed<rCommands && nextLayer<=currentLayer)
		{
			feX("DrawCached::drawOutput",
					"layer processing failed, completed %d/%d layer %d->%d",
					completed,rCommands,currentLayer,nextLayer);
		}
		currentLayer=nextLayer;
	}

	if(a_clearQueue)
	{
		rCommands=0;
	}

#if FE_DCH_CACHE_DEBUG
		feLog("DrawCached::drawOutput queue %d done\n",queue);
#endif
}

void DrawCached::drawPoints(const SpatialVector *vertex,
		const SpatialVector *normal,U32 vertices,
		BWORD multicolor,const Color *color,
		sp<DrawBufferI> spDrawBuffer)
{
#if FE_DCH_CACHE_DEBUG
	feLog("DrawCached::drawPoints\n");
#endif
	drawGeometry(Command::e_points,
			vertex,normal,NULL,NULL,NULL,vertices,e_discrete,multicolor,color,
			NULL,NULL,NULL,0,spDrawBuffer);
}

void DrawCached::drawLines(const SpatialVector *vertex,
		const SpatialVector *normal,U32 vertices,
		StripMode strip,BWORD multicolor,const Color *color,
		const Vector3i *element,U32 elementCount,
		sp<DrawBufferI> spDrawBuffer)
{
#if FE_DCH_CACHE_DEBUG
	feLog("DrawCached::drawLines\n");
#endif
	//* TODO pass along element and elementCount, if set
	drawGeometry(Command::e_lines,
			vertex,normal,NULL,NULL,NULL,vertices,strip,multicolor,color,
			NULL,NULL,element,elementCount,spDrawBuffer);
}

void DrawCached::drawTriangles(const SpatialVector *vertex,
	const SpatialVector *normal,const Vector2f *texture,U32 vertices,
	StripMode strip,BWORD multicolor,const Color *color,
	const Array<I32>* hullPointMap,const Array<Vector4i>* hullFacePoint,
	sp<DrawBufferI> spDrawBuffer)
{
#if FE_DCH_CACHE_DEBUG
	feLog("DrawCached::drawTriangles\n");
#endif
	drawGeometry(Command::e_triangles,
			vertex,normal,texture,NULL,NULL,vertices,strip,multicolor,color,
			hullPointMap,hullFacePoint,NULL,0,spDrawBuffer);
}

void DrawCached::drawRectangles(const SpatialVector *vertex,U32 vertices,
		BWORD multicolor,const Color *color)
{
#if FE_DCH_CACHE_DEBUG
	feLog("DrawCached::drawRectangles\n");
#endif
	drawGeometry(Command::e_rectangles,
			vertex,NULL,NULL,NULL,NULL,vertices,e_discrete,multicolor,color,
			NULL,NULL,NULL,0,sp<Component>(NULL));
}

void DrawCached::drawSphere(const SpatialTransform &transform,
		const SpatialVector *scale,const Color &color)
{
#if FE_DCH_CACHE_DEBUG
	feLog("DrawCached::drawSphere\n");
#endif

	FEASSERT(m_inputQueue<m_queues);

	CommandQueue& rCommandQueue=m_pCommandQueue[m_inputQueue];
	U32& rCommands=rCommandQueue.m_commands;
	Command& rCommand=rCommandQueue.access(rCommands++);

	rCommand.m_spDrawMode=drawMode();

#if FE_DCH_CACHE_DEBUG
	feLog("[35mDrawCached::drawSphere queue %d command %d layer %d[0m\n",
			m_inputQueue,rCommands-1,rCommand.layer());
#endif

	adoptView(rCommand,view());
	rCommand.m_projection=view()->cameraDisplay()->projection();
	rCommand.m_instruction=Command::e_sphere;
	rCommand.m_stripMode=e_discrete;
	rCommand.m_multicolor=FALSE;
	rCommand.m_vertices=1;
	rCommand.m_spImageI=NULL;

	rCommand.m_pVertex=rCommand.m_localBuffer.m_pVertex;
	transform.copy16(&rCommand.m_pVertex[0][0]);

	if(scale)
	{
		rCommand.m_pVertex[6]= *scale;
	}
	else
	{
		set(rCommand.m_pVertex[6],1.0f,1.0f,1.0f);
	}

	rCommand.m_pNormal=NULL;

	rCommand.m_pColor=rCommand.m_localBuffer.m_pColor;
	rCommand.m_pColor[0]=color;

	rCommand.m_pTexture=NULL;
}

void DrawCached::drawCylinder(const SpatialTransform &transform,
		const SpatialVector *scale,Real baseScale,const Color &color,
		U32 slices)
{
#if FE_DCH_CACHE_DEBUG
	feLog("DrawCached::drawCylinder\n");
#endif

	FEASSERT(m_inputQueue<m_queues);

	CommandQueue& rCommandQueue=m_pCommandQueue[m_inputQueue];
	U32& rCommands=rCommandQueue.m_commands;
	Command& rCommand=rCommandQueue.access(rCommands++);

	rCommand.m_spDrawMode=drawMode();

#if FE_DCH_CACHE_DEBUG
	feLog("[35mDrawCached::drawCylinder queue %d command %d layer %d[0m\n",
			m_inputQueue,rCommands-1,rCommand.layer());
#endif

	adoptView(rCommand,view());
	rCommand.m_projection=view()->cameraDisplay()->projection();
	rCommand.m_instruction=Command::e_cylinder;
	rCommand.m_stripMode=e_discrete;
	rCommand.m_multicolor=FALSE;
	rCommand.m_vertices=1;
	rCommand.m_spImageI=NULL;

	rCommand.m_pVertex=rCommand.m_localBuffer.m_pVertex;
	transform.copy16(&rCommand.m_pVertex[0][0]);

	if(scale)
	{
		rCommand.m_pVertex[6]= *scale;
	}
	else
	{
		set(rCommand.m_pVertex[6],1.0f,1.0f,1.0f);
	}
	rCommand.m_pVertex[7][0]=baseScale;
	rCommand.m_pVertex[7][1]=slices;

	rCommand.m_pNormal=NULL;

	rCommand.m_pColor=rCommand.m_localBuffer.m_pColor;
	rCommand.m_pColor[0]=color;

	rCommand.m_pTexture=NULL;
}

void DrawCached::drawCylinders(const SpatialTransform* transform,
		const SpatialVector* scale,const Real *baseScale,const U32* slices,
		U32 cylinders,BWORD multicolor,const Color* color)
{
#if FE_DCH_CACHE_DEBUG
	feLog("DrawCached::drawCylinders\n");
#endif

	// store scale in normal
	drawGeometry(Command::e_cylinders,(SpatialVector*)transform,scale,NULL,
			baseScale,slices,cylinders,e_discrete,multicolor,color,
			NULL,NULL,NULL,0,sp<Component>(NULL));
}

void DrawCached::drawRaster(sp<ImageI> spImageI,const SpatialVector& location)
{
#if FE_DCH_CACHE_DEBUG
	feLog("DrawCached::drawRaster\n");
#endif

	FEASSERT(m_inputQueue<m_queues);

	CommandQueue& rCommandQueue=m_pCommandQueue[m_inputQueue];
	U32& rCommands=rCommandQueue.m_commands;
	Command& rCommand=rCommandQueue.access(rCommands++);

	rCommand.m_spDrawMode=drawMode();

#if FE_DCH_CACHE_DEBUG
	feLog("[35mDrawCached::drawRaster queue %d command %d layer %d[0m\n",
			m_inputQueue,rCommands-1,rCommand.layer());
#endif

	adoptView(rCommand,view());
	rCommand.m_projection=view()->cameraDisplay()->projection();
	rCommand.m_instruction=Command::e_raster;
	rCommand.m_stripMode=e_discrete;
	rCommand.m_multicolor=FALSE;
	rCommand.m_vertices=1;
	rCommand.m_spImageI=spImageI;

	rCommand.m_pVertex=rCommand.m_localBuffer.m_pVertex;
	rCommand.m_pVertex[0]=location;

	rCommand.m_pNormal=NULL;

	rCommand.m_pColor=NULL;

	rCommand.m_pTexture=NULL;
}

void DrawCached::draw(cp<DrawableI> cpDrawableI,const Color* color,
	sp<DrawBufferI> spDrawBuffer)
{
	drawTransformed(SpatialTransform::identity(),cpDrawableI,color,
			spDrawBuffer);
}

void DrawCached::drawTransformed(const SpatialTransform& transform,
	cp<DrawableI> cpDrawableI,const Color* color,sp<DrawBufferI> spDrawBuffer)
{
	if(cpDrawableI.isNull())
	{
		return;
	}

#if FE_DCH_CACHE_DEBUG
	feLog("DrawCached::drawTransformed\n");
#endif

	FEASSERT(m_inputQueue<m_queues);

	CommandQueue& rCommandQueue=m_pCommandQueue[m_inputQueue];
	U32& rCommands=rCommandQueue.m_commands;
	Command& rCommand=rCommandQueue.access(rCommands++);

#if FE_DCH_CACHE_DEBUG
	feLog("[35mDrawCached::draw queue %d command %d layer %d[0m\n",
			m_inputQueue,rCommands-1,rCommand.layer());
	feLog("transform\n%s\n",c_print(transform));
#endif

	adoptView(rCommand,view());
	rCommand.m_spDrawMode=drawMode();
	rCommand.m_spDrawBuffer=spDrawBuffer;
	rCommand.m_projection=view()->cameraDisplay()->projection();
	rCommand.m_instruction=Command::e_drawable;
	rCommand.m_stripMode=e_discrete;
	rCommand.m_multicolor=FALSE;
	rCommand.m_vertices=1;
	rCommand.m_cpDrawableI=cpDrawableI;
	rCommand.m_cpDrawableI.protect();

	rCommand.m_pVertex=rCommand.m_localBuffer.m_pVertex;
	transform.copy16(&rCommand.m_pVertex[0][0]);

	rCommand.m_pNormal=NULL;

	if(color)
	{
		rCommand.m_pColor=rCommand.m_localBuffer.m_pColor;
		rCommand.m_pColor[0]= *color;
	}
	else
	{
		rCommand.m_pColor=NULL;
	}

	rCommand.m_pTexture=NULL;
}

void DrawCached::drawAlignedText(const SpatialVector &location,
		const String text,const Color &color)
{
	FEASSERT(m_inputQueue<m_queues);

	CommandQueue& rCommandQueue=m_pCommandQueue[m_inputQueue];
	U32& rCommands=rCommandQueue.m_commands;
	Command& rCommand=rCommandQueue.access(rCommands++);

	rCommand.m_spDrawMode=drawMode();

#if FE_DCH_CACHE_DEBUG
	feLog("DrawCached::drawAlignedText queue %d command %d layer %d"
			" string \"%s\"\n",
			m_inputQueue,rCommands-1,rCommand.layer(),text.c_str());
#endif

	adoptView(rCommand,view());
	rCommand.m_projection=view()->cameraDisplay()->projection();
	rCommand.m_instruction=Command::e_text;
	rCommand.m_stripMode=e_discrete;
	rCommand.m_multicolor=FALSE;
	rCommand.m_vertices=1;
	rCommand.m_spImageI=NULL;

	rCommand.m_pVertex=rCommand.m_localBuffer.m_pVertex;
	rCommand.m_pVertex[0]=location;

	rCommand.m_pNormal=NULL;

	rCommand.m_pColor=rCommand.m_localBuffer.m_pColor;
	rCommand.m_pColor[0]=color;

	rCommand.m_pTexture=rCommand.m_localBuffer.m_pTexture;
	char* buffer=reinterpret_cast<char*>(rCommand.m_pTexture);
	const FE_UWORD length=fe::minimum(text.length()+1,
			U32(DCH_MAX_VERTS*sizeof(Vector2f)));
	strncpy(buffer,text.c_str(),length);
	buffer[length-1]=0;
}

void DrawCached::drawGeometry(Command::Instruction instruction,
		const SpatialVector *vertex,const SpatialVector *normal,
		const Vector2f *texture,const Real* auxReal,const U32* auxU32,
		U32 vertices,StripMode strip,BWORD multicolor,const Color *color,
		const Array<I32>* hullPointMap,const Array<Vector4i>* hullFacePoint,
		const Vector3i *element,U32 elementCount,
		sp<DrawBufferI> spDrawBuffer)
{
	FEASSERT(m_inputQueue<m_queues);

	CommandQueue& rCommandQueue=m_pCommandQueue[m_inputQueue];
	U32& rCommands=rCommandQueue.m_commands;

	const U32 grain=m_grain[instruction];
	const U32 oversize=m_oversize[instruction];
	const U32 limit=(DCH_MAX_VERTS/(grain*oversize))*grain;
	const U32 groups=(vertices>=limit && strip==DrawI::e_discrete)?
			(vertices-1)/limit+1: 1;

	//* TODO split elements
	if(groups>1 && element)
	{
		feLog("DrawCached::drawGeometry splitting elements not implemented\n");
		element=NULL;
		elementCount=0;
	}

	// TODO what if not discrete and too many verts

#if FE_DCH_CACHE_DEBUG
	feLog("[35mDrawCached::drawGeometry queue %d command %d layer %d"
			" vertices %d strip %d multicolor %d\n"
			" grain %d limit %d groups %d\n"
			" vertex %p normal %p color %p texture %p[0m\n",
			m_inputQueue,rCommands,drawMode()->layer(),
			vertices,strip,multicolor,
			grain, limit, groups,
			vertex,normal,color,texture);
#endif

	for(U32 group=0;group<groups;group++)
	{
		Command& rCommand=rCommandQueue.access(rCommands++);

		U32 offset=limit*group;
		U32 subvertices=(group==groups-1)? vertices-offset: limit;

#if FE_DCH_CACHE_DEBUG
		feLog("  command %d group %d offset %d subvertices %d ptr=%p %p\n",
				rCommands-1,group,offset,subvertices,
				vertex+offset,color+offset);
#endif

		//* TODO continue strip in another group
		if(subvertices>limit)
		{
			feLog("DrawCached::drawGeometry clipping long strip %d to %d\n",
					subvertices,limit);

			subvertices=limit;
		}

		adoptView(rCommand,view());
		rCommand.m_spDrawMode=drawMode();
		rCommand.m_spDrawBuffer=spDrawBuffer;
		rCommand.m_pHullPointMap=hullPointMap;
		rCommand.m_pHullFacePoint=hullFacePoint;
		rCommand.m_pHullPointMap=hullPointMap;
		rCommand.m_elementCount=elementCount;
		rCommand.m_projection=view()->cameraDisplay()->projection();
		rCommand.m_instruction=instruction;
		rCommand.m_stripMode=strip;
		rCommand.m_multicolor=multicolor;
		rCommand.m_vertices=subvertices;
		rCommand.m_spImageI=NULL;

		rCommand.m_pVertex=rCommand.m_localBuffer.m_pVertex;
		memcpy(rCommand.m_pVertex,vertex+oversize*offset,
				subvertices*oversize*sizeof(SpatialVector));

		if(!normal)
		{
			rCommand.m_pNormal=NULL;
		}
		else
		{
			rCommand.m_pNormal=rCommand.m_localBuffer.m_pNormal;
			memcpy(rCommand.m_pNormal,normal+offset,
					subvertices*sizeof(SpatialVector));
		}

		if(!color)
		{
			rCommand.m_pColor=NULL;
		}
		else if (multicolor)
		{
			rCommand.m_pColor=rCommand.m_localBuffer.m_pColor;
			memcpy(rCommand.m_pColor,color+offset,
					subvertices*sizeof(Color));
		}
		else
		{
			rCommand.m_pColor=rCommand.m_localBuffer.m_pColor;
			rCommand.m_pColor[0]=color[0];
		}

		if(!texture)
		{
			rCommand.m_pTexture=NULL;
		}
		else
		{
			rCommand.m_pTexture=rCommand.m_localBuffer.m_pTexture;
			memcpy(rCommand.m_pTexture,texture+offset,
					subvertices*sizeof(Vector2f));
		}

		if(!element)
		{
			rCommand.m_pElement=NULL;
		}
		else
		{
			rCommand.m_pElement=rCommand.m_localBuffer.m_pElement;
			memcpy(rCommand.m_pElement,element,
					elementCount*sizeof(Vector3i));
		}

		if(rCommand.m_instruction==Command::e_cylinders)
		{
			rCommand.m_pTexture=rCommand.m_localBuffer.m_pTexture;

			CylinderOverlay* pCylinderOverlay=
					(CylinderOverlay*)rCommand.m_pTexture;
			Real* pBaseScale=pCylinderOverlay->m_pBaseScale;
			U32* pSlices=pCylinderOverlay->m_pSlices;

			memcpy(pBaseScale,auxReal+offset,subvertices*sizeof(Real));
			memcpy(pSlices,auxU32+offset,subvertices*sizeof(U32));
		}
	}
}

void DrawCached::clearCameraDisplayMap(void)
{
	m_cameraDisplayMap.clear();
	m_spWindow=NULL;
}

void DrawCached::adoptView(DrawCached::Command& rCommand,sp<ViewI> spViewI)
{
	sp<CameraDisplay> spCameraDisplay=spViewI->cameraDisplay();

	sp<CameraDisplay> spMappedMode=m_cameraDisplayMap[spCameraDisplay];
	if(!spMappedMode.isValid())
	{
		spMappedMode=new CameraDisplay();
		m_cameraDisplayMap[spCameraDisplay]=spMappedMode;
	}

	rCommand.m_spCameraDisplay=spMappedMode;

	if(spCameraDisplay.isValid())
	{
		spMappedMode->copy(spCameraDisplay);
	}

	m_spWindow=spViewI->window();
}

DrawCached::Command& DrawCached::Command::operator=(
		const DrawCached::Command& rCommand)
{
	m_spCameraDisplay=rCommand.m_spCameraDisplay;
	m_spDrawMode=rCommand.m_spDrawMode;
	m_spDrawBuffer=rCommand.m_spDrawBuffer;
	m_spImageI=rCommand.m_spImageI;
	m_cpDrawableI=rCommand.m_cpDrawableI;
	m_projection=rCommand.m_projection;
	m_instruction=rCommand.m_instruction;
	m_stripMode=rCommand.m_stripMode;
	m_multicolor=rCommand.m_multicolor;
	m_vertices=rCommand.m_vertices;
	m_elementCount=rCommand.m_elementCount;

	if(rCommand.m_pVertex)
	{
		m_pVertex=m_localBuffer.m_pVertex;
		memcpy(m_pVertex,rCommand.m_pVertex,
				DCH_MAX_VERTS*sizeof(SpatialVector));
	}
	else
	{
		m_pVertex=NULL;
	}

	if(rCommand.m_pNormal)
	{
		m_pNormal=m_localBuffer.m_pNormal;
		memcpy(m_pNormal,rCommand.m_pNormal,
				DCH_MAX_VERTS*sizeof(SpatialVector));
	}
	else
	{
		m_pNormal=NULL;
	}

	if(rCommand.m_pColor)
	{
		m_pColor=m_localBuffer.m_pColor;
		memcpy(m_pColor,rCommand.m_pColor,
				DCH_MAX_VERTS*sizeof(Color));
	}
	else
	{
		m_pColor=NULL;
	}

	if(rCommand.m_pTexture)
	{
		m_pTexture=m_localBuffer.m_pTexture;
		memcpy(m_pTexture,rCommand.m_pTexture,
				DCH_MAX_VERTS*sizeof(Vector2f));
	}
	else
	{
		m_pTexture=NULL;
	}

	if(rCommand.m_pElement)
	{
		m_pElement=m_localBuffer.m_pElement;
		memcpy(m_pElement,rCommand.m_pElement,
				DCH_MAX_VERTS*sizeof(Vector2f));
	}
	else
	{
		m_pElement=NULL;
	}

	return *this;
}

} // namespace
