/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "lua.pmh"

namespace fe
{

LuaHandler::LuaHandler(void)
{
	m_pLuaState = NULL;
	m_name = "unnamed";
}

LuaHandler::~LuaHandler(void)
{
	flush();
#if FE_COUNTED_TRACK
	U32 size=m_trackArray.size();
	for(U32 m=0;m<size;m++)
	{
		Counted::deregisterRegion(m_trackArray[m]);
	}
#endif
	if(m_pLuaState) { lua_close(m_pLuaState); }
}

void LuaHandler::initialize(void)
{
	m_pLuaState = lua_newstate(lua_alloc, NULL);
	luaL_reg lualibs[] = {
		{ "base",       luaopen_base },
		{ "table",      luaopen_table },
		{ "io",         luaopen_io },
		{ "string",     luaopen_string },
		{ "math",       luaopen_math },
		{ "debug",      luaopen_debug },
		{ "loadlib",    luaopen_package },
		{ NULL,         NULL }
		};
		luaL_reg *lib;


		for (lib = lualibs; lib->func != NULL; lib++)
		{
			lua_pushcfunction(m_pLuaState, lib->func);
			lua_pushstring(m_pLuaState, lib->name);
			lua_call(m_pLuaState, 1, 0);
			lua_settop(m_pLuaState, 0);
		}

	m_spContext = new LuaContext(registry()->master(), m_pLuaState);

	sp<TypeMaster> spTM=registry()->master()->typeMaster();

	dispatch<String>("loadString");

	dispatch<String>("setRecord");
	dispatch<Record>("setRecord");
}

void LuaHandler::alias(String a_aliasName,String a_trueName)
{
	if(m_spContext.isValid())
	{
		m_spContext->alias(a_aliasName,a_trueName);
	}
}

bool LuaHandler::preloadString(const String &text)
{
	m_preload=text;
	return true;
}

bool LuaHandler::loadFile(const String &filename)
{
	m_name = filename;
	std::ifstream file;
	file.open(filename.c_str());
	if(!file) { return false; }
	file.seekg (0, std::ios::end);
	int length = file.tellg();
	file.seekg (0, std::ios::beg);

	char *buffer = new char [length+1];
	memset((void *)buffer, 0, length+1); // make valgrind happy

	file.read (buffer,length);

	m_chunk = buffer;

	delete [] buffer;
	file.close();

	return compile();
}

bool LuaHandler::loadString(const String &text)
{
	m_name = "loaded string";
	m_chunk = text;
	return compile();
}

bool LuaHandler::compile(void)
{
	m_source=m_preload+m_chunk;

	int result;
	result = luaL_loadbuffer(m_pLuaState, m_source.c_str(),
			m_source.length(), m_source.c_str());
	if(result != 0)
	{
		feLog("LUA syntax error: %s \n", lua_tostring(m_pLuaState, -1));
		return false;
	}

	lua_setglobal(m_pLuaState, m_name.c_str());

	return true;
}

bool LuaHandler::execute(const String &a_fnName)
{
	lua_pushcfunction(m_pLuaState, LuaContext::debug);
	lua_pushstring(m_pLuaState, a_fnName.c_str());
	lua_gettable(m_pLuaState, LUA_GLOBALSINDEX);
	if(lua_isnil(m_pLuaState, -1))
	{
		lua_pop(m_pLuaState, 1);
		lua_pop(m_pLuaState, 1);
		return false;
	}
	int result = lua_pcall(m_pLuaState, 0, 0, -2);
	if(result)
	{
		String err = lua_tostring(m_pLuaState, -1);
		lua_pop(m_pLuaState, 1);
		feLog(err.c_str());
		return false;
	}
	lua_pop(m_pLuaState, 1);
	return true;
}


void LuaHandler::set(const String &key, Record r_value)
{
	t_lua_record::push(m_spContext.raw(), r_value);
	lua_setglobal(m_pLuaState, key.c_str());

// TODO restore for RecordAV
#if FALSE //FE_COUNTED_TRACK
	Record& record=r_value;
	if(record.isValid())
	{
		void* data=record.data();
		U32 size=record.layout()->size();

		Counted::registerRegion(data,size,
				"LuaHandler:Record "+record.layout()->name());
		Counted::trackReference(data,dynamic_cast<Counted*>(this),
				"LuaHandler");

		m_trackArray.push_back(data);
	}
#endif
}

void LuaHandler::set(const String &key, sp<RecordArray> ra_value)
{
	t_lua_array::push(m_spContext.raw(), ra_value);
	lua_setglobal(m_pLuaState, key.c_str());
}

void LuaHandler::set(const String &key, const String &value)
{
	lua_pushstring(m_pLuaState, value.c_str());
	lua_setglobal(m_pLuaState, key.c_str());
}

void LuaHandler::set(const String &key, const Real &value)
{
	lua_pushnumber(m_pLuaState, value);
	lua_setglobal(m_pLuaState, key.c_str());
}

void LuaHandler::set(const String &key, Vector3 &value)
{
	t_lua_vector3::push(m_spContext.raw(), &value);
	lua_setglobal(m_pLuaState, key.c_str());
}

void LuaHandler::set(const String &key, sp<Layout> l_value)
{
	t_lua_layout::push(m_spContext.raw(), l_value);
	lua_setglobal(m_pLuaState, key.c_str());
}

void LuaHandler::set(const String &key, sp<Component> spValue)
{
	lua_pushstring(m_pLuaState, key.c_str());
	LuaComponentObject::pushComponent(m_pLuaState, spValue);
	lua_settable(m_pLuaState, LUA_GLOBALSINDEX);
	//lua_setglobal(m_pLuaState, key.c_str());
}

void LuaHandler::set(const String &key, sp<Scope> spValue)
{
	t_lua_scope::push(m_spContext.raw(), spValue);
	lua_setglobal(m_pLuaState, key.c_str());
}

bool LuaHandler::execute(void)
{
	lua_pushcfunction(m_pLuaState, LuaContext::debug);

	lua_getglobal(m_pLuaState, m_name.c_str());

	int result = lua_pcall(m_pLuaState, 0, 0, -2);
	if(result)
	{
		String err = lua_tostring(m_pLuaState, -1);
		lua_pop(m_pLuaState, 1);
		feLog(err.c_str());
		return false;
	}
	lua_pop(m_pLuaState, 1);

#if 0
	// force garbage collection
	lua_setgcthreshold(m_pLuaState,0);
#endif

	return true;
}

void LuaHandler::flush(void)
{
	lua_gc(m_pLuaState,LUA_GCCOLLECT,0);
}

void LuaHandler::handle(Record& r_sig)
{
	set("r_sig", r_sig);
	execute();
}

bool LuaHandler::call(const String &a_name, Array<Instance> a_argv)
{
	if(a_name == FE_DISPATCH("loadString", "[string]"))
	{
		String text = a_argv[0].cast<String>();
		if(!loadString(text))
		{
			feX("LuaHandler::call",
				"lua script error");
		}
	}

	if(a_name == FE_DISPATCH("setRecord", "[name] [record]"))
	{
		String name = a_argv[0].cast<String>();
		Record record = a_argv[1].cast<Record>();
		set(name, record);
	}

	return true;
}



} /* namsepace */

