/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#ifndef __surface_QuadTree_h__
#define __surface_QuadTree_h__

namespace fe
{

/**************************************************************************//**
	@brief Triangular storage using divisions by quarters

	@ingroup surface
*//***************************************************************************/
class FE_DL_EXPORT QuadTree:
	public FlatTree,
	public ObjectSafeShared<QuadTree>
{
	public:
				QuadTree(void);
virtual			~QuadTree(void)												{}

virtual	void	setRefinement(U32 a_refinement)	{ m_subdivMax=a_refinement; }

virtual	void	populate(
						const Vector3i* a_pElementArray,
						const SpatialVector* a_pVertexArray,
						const SpatialVector* a_pNormalArray,
						const Vector2* a_pUVArray,
						const Color* a_pColorArray,
						const I32* a_pTriangleIndexArray,
						const I32* a_pPointIndexArray,
						const I32* a_pPrimitiveIndexArray,
						const I32* a_pPartitionIndexArray,
						U32 a_primitives,U32 a_vertices,
						const SpatialVector& a_center,F32 a_radius);
virtual	Real	nearestPoint(const SpatialVector& a_origin,
						Real a_maxDistance,BWORD a_anyHit,U32 a_hitLimit,
						const sp<PartitionI>& a_rspPartition,
						Array< sp<SpatialTreeI::Hit> >& a_rHitArray)
						const;
virtual	Real	rayImpact(const SpatialVector& a_origin,
						const SpatialVector& a_direction,
						Real a_maxDistance,BWORD a_anyHit,U32 a_hitLimit,
						const sp<PartitionI>& a_rspPartition,
						Array< sp<SpatialTreeI::Hit> >& a_rHitArray)
						const;

virtual	void	draw(sp<DrawI> a_spDrawI,const fe::Color* a_pColor) const;

		Real	nearestTo(BWORD a_mutable,BWORD a_directional,
						const SpatialVector& a_origin,
						const SpatialVector& a_direction,
						Real a_maxDistance,BWORD a_anyHit,U32 a_hitLimit,
						const sp<PartitionI>& a_rspPartition,
						Array< sp<SpatialTreeI::Hit> >& a_rHitArray,
						U32 a_subdivMax) const;

		void	age(void);

	class Element
	{
		public:
					Element(void):
						m_face(0),
						m_triangleIndex(-1),
						m_primitiveIndex(-1),
						m_partitionIndex(-1),
						m_sphereIndex(0),
						m_sortValue(0.0)									{}
			I32		m_face;
			I32		m_triangleIndex;
			I32		m_primitiveIndex;
			I32		m_partitionIndex;
			U32		m_sphereIndex;
			F32		m_sortValue;

	static	int		compare(const void* a_pVoid1,const void* a_pVoid2);
	};

	class FE_DL_EXPORT BaseNode: public DAGNode
	{
		public:
			U32				m_level;
			Vector4			m_boundSphere;
			String			m_debugText;

							BaseNode(void):
								m_level(0)
							{
								setName("QuadTree::BaseNode");
								set(m_boundSphere);
							}
	virtual					~BaseNode(void)									{}

			void			subdivide(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									const Vector2* a_pUVArray,
									const Element* a_pElements,
									U32 a_elementIndex,U32 a_elementCount,
									U32 a_level,U32 a_maxLevel,
									U32 a_timeStamp);

	virtual	void			subdivideOne(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									const Vector2* a_pUVArray,
									U32 a_triangleIndex,
									U32 a_level,U32 a_maxLevel,
									U32 a_timeStamp)						=0;
	virtual	void			subdivideAll(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									const Vector2* a_pUVArray,
									U32 a_level,U32 a_maxLevel,
									U32 a_timeStamp)						=0;

			void			draw(sp<DrawI> a_spDrawI,
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									const Color* a_pColor,
									U32 a_level) const;
	static	void			calcFrames(
									SpatialTransform& a_rTransform1,
									SpatialTransform& a_rTransform2,
									const SpatialVector& a_rVertex1,
									const SpatialVector& a_rNormal1,
									const SpatialVector& a_rVertex2,
									const SpatialVector& a_rNormal2);
			Real			separation(const SpatialVector& a_origin) const;

			void			nearestTo(BWORD a_mutable,BWORD a_directional,
									const SpatialVector& a_origin,
									const SpatialVector& a_direction,
									Real a_maxDistance,BWORD a_anyHit,
									U32 a_hitLimit,
									const sp<PartitionI>& a_rspPartition,
									Array< sp<SpatialTreeI::Hit> >&
									a_rHitArray,
									CountedPool<Hit>& a_rHitPool,
									U32 a_subdivMax,
									SurfaceI::Accuracy a_accuracy,
									const Vector3i* a_pElementArray,
									const I32* a_pPointIndexArray,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									const Vector2* a_pUVArray,
									const Color* a_pColorArray,
									const Vector4* a_pSphereArray,
									U32 a_timeStamp) const;

	virtual I32				face(U32 a_index) const							=0;
	virtual	U32				faceCount(void) const							=0;
	virtual
	const I32*				facePointIndex(
									const Vector3i* a_pElementArray,
									const I32* a_pPointIndexArray,
									U32 a_index) const						=0;
	virtual
	const SpatialVector*	faceVertex(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									U32 a_index) const						=0;
	virtual
	const SpatialVector*	faceNormal(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pNormalArray,
									U32 a_index) const						=0;
	virtual
	const Vector2*			faceUV(
									const Vector3i* a_pElementArray,
									const Vector2* a_pUVArray,
									U32 a_index) const						=0;
	virtual
	const Color*			faceColor(
									const Vector3i* a_pElementArray,
									const Color* a_pColorArray,
									U32 a_index) const						=0;

	virtual I32				triangleIndex(U32 a_index) const	{ return -1; }
	virtual I32				primitiveIndex(U32 a_index) const	{ return -1; }
	virtual I32				partitionIndex(U32 a_index) const	{ return -1; }
	virtual
	const Vector4			sphereData(U32 a_index,
									const Vector4* a_pSphereArray) const
							{	return Vector4(0.0f,0.0f,0.0f,0.0f); }

			void			prune(U32 a_timeStamp);
	};

	class FE_DL_EXPORT Node: public BaseNode
	{
		public:
			Element*		m_pElements;
			U32				m_elementCount;

							Node(void):
								m_pElements(NULL),
								m_elementCount(0)
							{
								setName("QuadTree::Node");
							}

	virtual					~Node(void)
							{	delete[] m_pElements; }

			void			refine(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									SpatialVector* a_pVertexScratch,
									const Vector3u& a_divisions,
									const Vector4* a_pSphereArray,
									U32 a_capacity);
			void			split(const Vector4* a_pSphereArray,U32 axis,
									Element*& a_rpElements,
									U32& a_rElementCount,U32 divisions,
									Element** a_rppElements,
									U32* a_rpElementCount);
			void			calcBounds(
									const SpatialVector* a_pVertexArray,
									const U32 a_vertexCount);
			void			calcBounds(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									SpatialVector* a_pVertexScratch);

	virtual	void			subdivideOne(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									const Vector2* a_pUVArray,
									U32 a_triangleIndex,
									U32 a_level,U32 a_maxLevel,U32 a_timeStamp);
	virtual	void			subdivideAll(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									const Vector2* a_pUVArray,
									U32 a_level,U32 a_maxLevel,
									U32 a_timeStamp);

	virtual I32				face(U32 a_index) const
							{	FEASSERT(a_index<m_elementCount);
								return m_pElements[a_index].m_face; }
	virtual	U32				faceCount(void) const
							{	return m_elementCount; }
	virtual
	const I32*				facePointIndex(
									const Vector3i* a_pElementArray,
									const I32* a_pPointIndexArray,
									U32 a_index) const
							{
								const U32 startIndex=
										a_pElementArray[face(a_index)][0];
								return &a_pPointIndexArray[startIndex];
							}
	virtual
	const SpatialVector*	faceVertex(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									U32 a_index) const
							{
								const U32 startIndex=
										a_pElementArray[face(a_index)][0];
								return &a_pVertexArray[startIndex];
							}
	virtual
	const SpatialVector*	faceNormal(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pNormalArray,
									U32 a_index) const
							{
								const U32 startIndex=
										a_pElementArray[face(a_index)][0];
								return &a_pNormalArray[startIndex];
							}
	virtual
	const Vector2*			faceUV(
									const Vector3i* a_pElementArray,
									const Vector2* a_pUVArray,
									U32 a_index) const
							{
								const U32 startIndex=
										a_pElementArray[face(a_index)][0];
								return a_pUVArray? &a_pUVArray[startIndex]:
										NULL;
							}
	virtual
	const Color*			faceColor(
									const Vector3i* a_pElementArray,
									const Color* a_pColorArray,
									U32 a_index) const
							{
								const U32 startIndex=
										a_pElementArray[face(a_index)][0];
								return a_pColorArray?
										&a_pColorArray[startIndex]: NULL;
							}

	virtual I32				triangleIndex(U32 a_index) const
							{	FEASSERT(a_index<m_elementCount);
								return m_pElements[a_index].m_triangleIndex; }
	virtual I32				primitiveIndex(U32 a_index) const
							{	FEASSERT(a_index<m_elementCount);
								return m_pElements[a_index].m_primitiveIndex; }
	virtual I32				partitionIndex(U32 a_index) const
							{	FEASSERT(a_index<m_elementCount);
								return m_pElements[a_index].m_partitionIndex; }
	virtual
	const Vector4			sphereData(U32 a_index,
									const Vector4* a_pSphereArray) const
							{	FEASSERT(a_index<m_elementCount);
								return a_pSphereArray
										[m_pElements[a_index].m_sphereIndex]; }
	};

	class FE_DL_EXPORT SubNode: public BaseNode
	{
		public:
			Vector3i*		m_pSubdivPrimitiveArray;
			SpatialVector*	m_pSubdivVertexArray;
			SpatialVector*	m_pSubdivNormalArray;
			Vector2*		m_pSubdivUVArray;
			U32				m_serial;
			U32				m_subdivCount;
			U32				m_subdivLevel;
			U32				m_timeStamp;

							SubNode(void):
								m_pSubdivPrimitiveArray(NULL),
								m_pSubdivVertexArray(NULL),
								m_pSubdivNormalArray(NULL),
								m_pSubdivUVArray(NULL),
								m_serial(0),
								m_subdivCount(0),
								m_subdivLevel(0),
								m_timeStamp(0)
							{
								setName("QuadTree::SubNode");
							}

	virtual					~SubNode(void)
							{
								delete[] m_pSubdivUVArray;
								delete[] m_pSubdivNormalArray;
								delete[] m_pSubdivVertexArray;
								delete[] m_pSubdivPrimitiveArray;
							}

	virtual	void			subdivideOne(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									const Vector2* a_pUVArray,
									U32 a_triangleIndex,
									U32 a_level,U32 a_maxLevel,U32 a_timeStamp);
	virtual	void			subdivideAll(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									const SpatialVector* a_pNormalArray,
									const Vector2* a_pUVArray,
									U32 a_level,U32 a_maxLevel,
									U32 a_timeStamp);

	virtual I32				face(U32 a_index) const	{ return a_index; }
	virtual	U32				faceCount(void) const
							{	return m_subdivCount/3; }
	virtual
	const I32*				facePointIndex(
									const Vector3i* a_pElementArray,
									const I32* a_pPointIndexArray,
									U32 a_index) const
							{
								return NULL;
							}
	const SpatialVector*	faceVertex(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pVertexArray,
									U32 a_index) const
//							{	FEASSERT((a_index+1)*a_grain<=m_subdivCount);
//								return &m_pSubdivVertexArray[a_index*a_grain]; }
							{
								const U32 startIndex=
										a_pElementArray[a_index][0];
								return &m_pSubdivVertexArray[startIndex];
							}

	virtual
	const SpatialVector*	faceNormal(
									const Vector3i* a_pElementArray,
									const SpatialVector* a_pNormalArray,
									U32 a_index) const
//							{	FEASSERT((a_index+1)*a_grain<=m_subdivCount);
//								return &m_pSubdivNormalArray[a_index*a_grain]; }
							{
								const U32 startIndex=
										a_pElementArray[a_index][0];
								return &m_pSubdivNormalArray[startIndex];
							}

	virtual
	const Vector2*			faceUV(
									const Vector3i* a_pElementArray,
									const Vector2* a_pUVArray,
									U32 a_index) const
//							{	FEASSERT((a_index+1)*a_grain<=m_subdivCount);
//								return m_pSubdivUVArray?
//										&m_pSubdivUVArray[a_index*a_grain]:
//										NULL; }
							{
								const U32 startIndex=
										a_pElementArray[a_index][0];
								return m_pSubdivUVArray?
										&m_pSubdivUVArray[startIndex]:
										NULL;
							}

	virtual
	const Color*			faceColor(
									const Vector3i* a_pElementArray,
									const Color* a_pColorArray,
									U32 a_index) const
							{
								//* TODO
								return NULL;
							}
	};

	class Child
	{
		public:
							Child(void): m_spNode(NULL),m_distance(0.0)		{}
			sp<BaseNode>	m_spNode;
			F32				m_distance;

	static	int				compare(const void* a_pVoid1,const void* a_pVoid2);
	};

		void				cacheSpheres(const SpatialVector& a_center,
									F32 a_radius);
		void				buildGraph(void);

		sp<Node>			m_spRootNode;
		sp<DrawMode>		m_spWireframe;

	protected:
		void				prune(void);
		void				lockSafe(void);
		void				unlockSafe(void);

		Vector3u			m_divisions;
		U32					m_capacity;
		U32					m_subdivMax;
		U32					m_ageStamp;
		U32					m_timeStamp;
};

} /* namespace */

#endif /* __surface_QuadTree_h__ */
