/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "vegetation/vegetation.pmh"

#define FE_TOP_SIGNAL_DEBUG		FALSE

using namespace fe;

TreeOp::TreeOp(void)
{
}

TreeOp::~TreeOp(void)
{
#if FE_TOP_SIGNAL_DEBUG
	Peeker peeker;
	if(m_spSignalerI.isValid())
	{
		m_spSignalerI->peek(peeker);
	}

	const String peek=peeker.str();
	if(!peek.empty())
	{
		feLog("\n~TreeOp\n%s\n",peek.c_str());
	}
#endif
}

void TreeOp::initialize(void)
{
	catalog<String>("icon")="FE_beta";

	catalog<String>("TreeName")="Aspen_Quaking";
	catalog<String>("TreeName","label")="Tree Name";
	catalog<String>("TreeName","hint")="Tree growth configuration"
			" (ignored when input geometry is provided).";

	catalog<SpatialVector>("TreeLocation");
	catalog<String>("TreeLocation","label")="Tree Location";
	catalog<String>("TreeLocation","hint")=
			"Displacement of trunk base from origin.";

	catalog<SpatialVector>("WindDirection");
	catalog<String>("WindDirection","label")="Wind Direction";

	catalog<String>("stickForm")="line";
	catalog<String>("stickForm","label")="Branch Form";
	catalog<String>("stickForm","choice:0")="none";
	catalog<String>("stickForm","label:0")="None";
	catalog<String>("stickForm","choice:1")="line";
	catalog<String>("stickForm","label:1")="Line";
	catalog<String>("stickForm","choice:2")="cylinder";
	catalog<String>("stickForm","label:2")="Cylinder";
	catalog<String>("stickForm","choice:2")="curve";
	catalog<String>("stickForm","label:2")="Curve";
	catalog<String>("stickForm","hint")=
			"Geometry used to output branch segments.";

	catalog<String>("leafForm")="none";
	catalog<String>("leafForm","label")="Leaf Form";
	catalog<String>("leafForm","choice:0")="none";
	catalog<String>("leafForm","label:0")="None";
	catalog<String>("leafForm","choice:1")="point";
	catalog<String>("leafForm","label:1")="Point";
	catalog<String>("leafForm","choice:2")="circle";
	catalog<String>("leafForm","label:2")="circle";
	catalog<String>("leafForm","hint")=
			"Geometry used to output each leaf.";

	catalog<bool>("YUp")=true;
	catalog<String>("YUp","label")="Y Is Up";
	catalog<String>("YUp","hint")=
			"Indicate that the Y axis points up instead of the Z axis.";

	catalog< sp<Component> >("Input Surface");
	catalog<bool>("Input Surface","optional")=true;

	catalog< sp<Component> >("Collider Surface");
	catalog<bool>("Collider Surface","optional")=true;

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
	catalog<bool>("Output Surface","temporal")=true;

	m_spSignalerI=registry()->create("SignalerI");
	sp<HandlerI> spSimPlant=registry()->create("*.SimPlant");
	sp<HandlerI> spDrawPlant=registry()->create("*.DrawPlant");
	if(!m_spSignalerI.isValid() || !spSimPlant.isValid() ||
			!spDrawPlant.isValid())
	{
		feX("TreeOp::initialize", "can't create components");
	}
	m_modifierArray.push_back(spSimPlant);
	m_modifierArray.push_back(spDrawPlant);

	m_spScope=registry()->master()->catalog()->catalogComponent(
			"Scope","PlantScope");
	FEASSERT(m_spScope.isValid());

	m_forestRV.bind(m_spScope);
	m_forestRV.createRecord();
	set(m_forestRV.effectorVelocity());
	set(m_forestRV.windSource());

	m_plantRV.bind(m_spScope);

	m_spSignalerI->insert(spSimPlant,m_forestRV.layout());
	m_spSignalerI->insert(spDrawPlant,m_forestRV.layout());
}

SpatialVector TreeOp::convertY(SpatialVector a_vector)
{
	return catalog<bool>("YUp")?
			SpatialVector(a_vector[0],-a_vector[2],a_vector[1]): a_vector;
}

SpatialVector TreeOp::unconvertY(SpatialVector a_vector)
{
	return catalog<bool>("YUp")?
			SpatialVector(a_vector[0],a_vector[2],-a_vector[1]): a_vector;
}

BWORD TreeOp::rebuild(sp<SurfaceAccessibleI>& a_rspSurfaceAccessible)
{
	feLog("TreeOp::rebuild\n");

	setName("TreeOp (skeleton)");

	sp<SurfaceAccessorI> spInputVertices;
	if(!access(spInputVertices,a_rspSurfaceAccessible,e_primitive,e_vertices))
	{
		return FALSE;
	}

	sp<SurfaceAccessorI> spInputPoints;
	if(!access(spInputPoints,a_rspSurfaceAccessible,e_point,e_generic))
	{
		return FALSE;
	}
	const U32 pointCount=spInputPoints->count();

	m_forestRV.plants()->clear();

	Record plant=m_plantRV.createRecord();
	m_plantRV.plantName()="<skeleton>";
	m_plantRV.location()=convertY(catalog<SpatialVector>("TreeLocation"));

	sp<PlantModelI> spPlantModelI=registry()->create("PlantModelI.Tree");
	m_plantRV.plantModel()=spPlantModelI;

	Array<I32> added(pointCount);
	for(U32 pointIndex=0;pointIndex<pointCount;pointIndex++)
	{
		added[pointIndex]=false;
	}

	const U32 primitiveCount=spInputVertices->count();
	for(U32 primitiveIndex=0;primitiveIndex<primitiveCount;primitiveIndex++)
	{
		const U32 subCount=spInputVertices->subCount(primitiveIndex);
		if(!subCount)
		{
			continue;
		}

		U32 fromIndex=spInputVertices->integer(primitiveIndex,0);
		FEASSERT(fromIndex<pointCount);

		SpatialVector fromPosition=convertY(
				spInputVertices->spatialVector(primitiveIndex,0));

		for(U32 subIndex=1;subIndex<subCount;subIndex++)
		{
			const U32 toIndex=spInputVertices->integer(primitiveIndex,subIndex);
			FEASSERT(toIndex<pointCount);

			SpatialVector toPosition=convertY(
					spInputVertices->spatialVector(primitiveIndex,subIndex));
			if(added[toIndex])
			{
				feLog("TreeOp::rebuild redundant segment"
						" from point %d to %d\n",fromIndex,toIndex);
				fromIndex=toIndex;
				fromPosition=toPosition;
				continue;
			}

			const U32 level=(primitiveIndex!=0);

			Real radiusFade=magnitude(toPosition)/30.0;
			if(radiusFade>1.0)
			{
				radiusFade=1.0;
			}
			const Real radius=0.10-0.09*radiusFade;

			spPlantModelI->addBranch(fromIndex,toIndex,level,radius,
					toPosition-fromPosition);

			fromIndex=toIndex;
			fromPosition=toPosition;
			added[toIndex]=true;
		}
	}

	spPlantModelI->prepare();

	m_forestRV.plants()->add(plant);

	m_plantRV.unbind();
	return TRUE;
}

void TreeOp::replace(sp<SurfaceAccessibleI>& a_rspSurfaceAccessible)
{
	sp<SurfaceAccessorI> spOutputPoint;
	if(!access(spOutputPoint,a_rspSurfaceAccessible,e_point,e_position))
	{
		return;
	}

	sp<RecordGroup> spRG=m_forestRV.plants();
	RecordGroup::iterator it=spRG->begin();
	if(it==spRG->end())
	{
		return;
	}
	sp<RecordArray> spRA=*it;
	if(!spRA->length())
	{
		return;
	}
	Record plant=spRA->getRecord(0);
	if(!plant.isValid())
	{
		return;
	}
	m_plantRV.bind(plant);

	sp<PlantModelI> spPlantModelI=m_plantRV.plantModel();

	const SpatialVector root=unconvertY(m_plantRV.location());

	const U32 numSticks=spPlantModelI->numSticks();
	const U32 pointCount=spOutputPoint->count();
	if(pointCount)
	{
		spOutputPoint->set(0,root);
	}
	for(U32 pointIndex=1;pointIndex<pointCount;pointIndex++)
	{
		SpatialVector tip=root+spOutputPoint->spatialVector(pointIndex);

		if(pointIndex<=numSticks)
		{
			const PlantModelI::Stick* pStick=
					spPlantModelI->stick(pointIndex-1);
			if(pStick)
			{
				const SpatialVector base=pStick->base();
				const SpatialVector span=pStick->span();
				tip=unconvertY(base+span);
			}
		}

		spOutputPoint->set(pointIndex,tip);
	}

	m_plantRV.unbind();
}

void TreeOp::regenerate(void)
{
	feLog("TreeOp::regenerate \"%s\"\n",m_treeName.c_str());

	setName(m_treeName);

	m_forestRV.plants()->clear();

	const String path=
			registry()->master()->catalog()->catalog<String>("path:media")+
			"/vegetation/"+m_treeName+".rg";

	sp<RecordGroup>	spRG=RecordView::loadRecordGroup(m_spScope,path);
	if(spRG.isNull())
	{
		feLog("TreeOp::regenerate failed to load:\n  \"%s\"\n",path.c_str());
		return;
	}

	RecordGroup::iterator it=spRG->begin();
	sp<RecordArray>& rspRA= *it;
	if(m_plantRV.plantName.check(rspRA))
	{
		m_forestRV.plants()->add(spRG);
	}
	else
	{
		Record plant=m_plantRV.createRecord();
		m_plantRV.plantName()=m_treeName;
		m_plantRV.location()=convertY(catalog<SpatialVector>("TreeLocation"));
		m_forestRV.plants()->add(plant);
	}

	m_plantRV.unbind();
}

void TreeOp::handle(Record& a_rSignal)
{
//	const Real frame=currentFrame(a_rSignal);
	const Real time=currentTime(a_rSignal);

//	feLog("TreeOp::handle frame %.6G time %.6G\n",frame,time);

	sp<SurfaceAccessibleI> spInputAccessible;
	access(spInputAccessible,"Input Surface",e_quiet);
	if(spInputAccessible.isValid())
	{
		m_treeName="";
		if(time<1e-6)
		{
			if(!rebuild(spInputAccessible))
			{
				return;
			}
		}
	}
	else
	{
		const String treeName=catalog<String>("TreeName");
		if(m_treeName!=treeName)
		{
			feLog("TreeOp::handle treeName changed \"%s\" -> \"%s\"\n",
					m_treeName.c_str(),treeName.c_str());
			m_treeName=treeName;
			regenerate();
		}
	}

	catalog<String>("summary")=m_treeName;

	sp<DrawI> spDrawI;
	if(!accessDraw(spDrawI,a_rSignal)) return;

	m_forestRV.upAxis()=catalog<bool>("YUp")? 1: 2;

	//* HACK always draw (for Arnold tests)
//	m_forestRV.drawI()=spInputAccessible.isValid()? sp<DrawI>(NULL): spDrawI;
	m_forestRV.drawI()=spDrawI;

	m_forestRV.stickForm()=catalog<String>("stickForm");
	m_forestRV.leafForm()=catalog<String>("leafForm");
	m_forestRV.deltaT()=(1.0f/24.0f);
	m_forestRV.windDirection()=catalog<SpatialVector>("WindDirection");

	sp<SurfaceI> spCollider;
	access(spCollider,"Collider Surface",e_quiet);

#if FALSE
	if(spCollider.isValid())
	{
		sp<SurfaceSphere> spSphere(new SurfaceSphere());
		spSphere->setCenter(convertY(spCollider->center()));
		spSphere->setRadius(spCollider->radius());

		spCollider=spSphere;
	}
#endif

	SpatialTransform colliderTransform;
	setIdentity(colliderTransform);

	if(catalog<bool>("YUp"))
	{
		colliderTransform.column(0)=SpatialVector(1.0, 0.0, 0.0);
		colliderTransform.column(1)=SpatialVector(0.0, 0.0, 1.0);
		colliderTransform.column(2)=SpatialVector(0.0,-1.0, 0.0);
		set(colliderTransform.translation());
	}

	m_forestRV.collider()=spCollider;
	m_forestRV.colliderTransform()=colliderTransform;

	sp<RecordGroup>& rspPlants=m_forestRV.plants();
	FEASSERT(rspPlants.isValid());

	//* update location, but only if it is an individual tree
	for(RecordGroup::iterator it=rspPlants->begin();it!=rspPlants->end();it++)
	{
		sp<RecordArray> spRA= *it;

		if(spRA->length()!=1)
		{
			continue;
		}

		if(!m_plantRV.plantModel.check(spRA))
		{
			continue;
		}

		m_plantRV.bind(spRA);
		while(m_plantRV.step())
		{
			m_plantRV.location()=
					convertY(catalog<SpatialVector>("TreeLocation"));
		}
	}

//	m_plantRV.bind(NULL);

	Record forest=m_forestRV.record();
	m_spSignalerI->signal(forest);

	if(spInputAccessible.isValid())
	{
		sp<SurfaceAccessibleI> spOutputAccessible;
		accessOutput(spOutputAccessible,a_rSignal);

		replace(spOutputAccessible);
	}

//	feLog("TreeOp::handle done\n");
}
