/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "fe/platform.h"

namespace fe
{

#if FE_LOCK_MEMORY
#if FE_SAFE_COUNTED_MUTEX
//* NOTE leak
I32 Memory::ms_mutexCount=0;
RecursiveMutex* Memory::ms_pMutex=NULL;
#else
RecursiveMutex Memory::ms_mutex;
#endif
#endif

void fe_memory_lock(void)
{
#if FE_LOCK_MEMORY
#if FE_SAFE_COUNTED_MUTEX
	if(Memory::ms_pMutex)
	{
		Memory::ms_pMutex->lock();
	}
	else if(!Memory::ms_mutexCount)
	{
		Memory::ms_mutexCount++;
		Memory::ms_pMutex=new RecursiveMutex;
		Memory::ms_pMutex->lock();
	}
#else
	ms_mutex.lock();
#endif
#endif
}

void fe_memory_unlock(void)
{
#if FE_LOCK_MEMORY
#if FE_SAFE_COUNTED_MUTEX
	if(Memory::ms_pMutex)
	{
		Memory::ms_pMutex->unlock();
	}
#else
	ms_mutex.unlock();
#endif
#endif
}

#ifdef FE_EXPORT_MEM_VIA_FUNC
FE_DL_EXPORT void* FE_CDECL ex_allocate(FE_UWORD byteCount)
{
	fe_memory_lock();

// NOTE OSX malloc is always 16-byte aligned
#if FE_MEM_ALIGNMENT && FE_OS!=FE_WIN32 && FE_OS!=FE_WIN64 && FE_OS!=FE_OSX
	void* pBuffer=NULL;
	int fail=posix_memalign(&pBuffer,FE_MEM_ALIGNMENT,byteCount);
//	printf("ex_allocate(%d) %p %d\n",int(byteCount),pBuffer,fail);
//	FEASSERT(pBuffer<(void*)0x200000000000);

#if FE_CHECK_HEAP
	if(!fail && (!gs_pHeapBase || gs_pHeapBase>pBuffer))
	{
		gs_pHeapBase=pBuffer;
	}
#elif FE_CHECK_NEW
	gs_pHeapBase=(void*)0x1;
#endif

	if(fail)
	{
		pBuffer=NULL;
	}

#elif FE_CODEGEN<=FE_DEBUG && (FE_OS==FE_WIN32 || FE_OS==FE_WIN64)
	void* pBuffer=malloc(byteCount);

#if FE_CHECK_HEAP
	if(!gs_pHeapBase || gs_pHeapBase>pBuffer)
	{
		gs_pHeapBase=pBuffer;
	}
#elif FE_CHECK_NEW
	gs_pHeapBase=(void*)0x1;
#endif

#else
	void* pBuffer=malloc(byteCount);
#endif

	fe_memory_unlock();

	return pBuffer;
}

FE_DL_EXPORT void * FE_CDECL ex_reallocate(void *pMemory, FE_UWORD byteCount)
{
	fe_memory_lock();

#if FE_MEM_ALIGNMENT && FE_OS!=FE_WIN32 && FE_OS!=FE_WIN64 && FE_OS!=FE_OSX
	void *pNew=realloc(pMemory, byteCount);
	if(pNew==pMemory)
	{
		fe_memory_unlock();

		return pNew;
	}

	void* pBuffer;
	int fail=posix_memalign(&pBuffer,FE_MEM_ALIGNMENT,byteCount);
//	printf("ex_reallocate(%d) %p %d\n",int(byteCount),pBuffer,fail);
	if(fail)
	{
		free(pNew);

		fe_memory_unlock();

		return NULL;
	}
	memcpy(pBuffer,pNew,byteCount);
	free(pNew);
#else
	void* pBuffer=realloc(pMemory, byteCount);
#endif

	fe_memory_unlock();

	return pBuffer;
}

FE_DL_EXPORT void  FE_CDECL ex_deallocate(void *pMemory)
{
	fe_memory_lock();

	free(pMemory);

	fe_memory_unlock();
}

#endif

#ifdef FE_EXPORT_MEMORY
#ifdef FE_EXPORT_MEM_VIA_FUNC
allocateFunction	*gs_pAllocateFunction	= ex_allocate;
deallocateFunction	*gs_pDeallocateFunction	= ex_deallocate;
reallocateFunction	*gs_pReallocateFunction	= ex_reallocate;
void*				gs_pHeapBase			= NULL;
I32					gs_newCheck				= 0;
#else
#if FE_COMPILER==FE_DMC
allocateFunction	*gs_pAllocateFunction	= malloc;
deallocateFunction	*gs_pDeallocateFunction	= free;
reallocateFunction	*gs_pReallocateFunction	= realloc;
void*				gs_pHeapBase			= NULL;
I32					gs_newCheck				= 0;
#else
FE_DL_EXPORT allocateFunction	*gs_pAllocateFunction	= malloc;
FE_DL_EXPORT deallocateFunction	*gs_pDeallocateFunction	= free;
FE_DL_EXPORT reallocateFunction	*gs_pReallocateFunction	= realloc;
FE_DL_EXPORT void*				gs_pHeapBase			= NULL;
FE_DL_EXPORT I32				gs_newCheck				= 0;

FE_DL_EXPORT void setAllocateFunction(allocateFunction *p_fn)
{
		gs_pAllocateFunction = p_fn;
}

FE_DL_EXPORT void setDeallocateFunction(deallocateFunction *p_fn)
{
		gs_pDeallocateFunction = p_fn;
}
#endif
#endif
#endif

} /* namespace */

#include <execinfo.h>
void logBacktrace(int size)
{
	const U32 backtrace_size = 16;
	void* stack[backtrace_size];
	I32 count=backtrace(stack,backtrace_size);
//	char** symbols=backtrace_symbols(stack,count);

	Dl_info info;

	for(I32 level=2;level<count;level++)
	{
		if(dladdr(stack[level],&info) && info.dli_sname)
		{
			//* fe::Mutex
			if(!strncmp(info.dli_sname,"_ZN2fe5Mutex",12))
			{
				return;
			}
			//* fe::data::AsciiReader
			if(!strncmp(info.dli_sname,"_ZN2fe4data11AsciiReader",24))
			{
				return;
			}
			//* fe::String::parse
			if(!strcmp(info.dli_sname,"_ZN2fe6String5parseES0_S0_b"))
			{
				return;
			}
			//* fe::String::vsPrintf
			if(!strcmp(info.dli_sname,
					"_ZN2fe6String8vsPrintfEPKcP13__va_list_tagRi"))
			{
				return;
			}
			//* create std::basic_string
			if(!strcmp(info.dli_sname,"_ZNSs4_Rep9_S_createEmmRKSaIcE"))
			{
				return;
			}
		}
	}

	fprintf(stderr,"*new %d*\n",size);

	for(I32 level=2;level<count;level++)
	{
		if(dladdr(stack[level],&info) && info.dli_sname)
		{
			fprintf(stderr,"  %s\n",info.dli_sname);

			if(!strcmp(info.dli_sname,"PyObject_Call") ||
					!strcmp(info.dli_sname,"fe_terminal_execute"))
			{
				return;
			}
		}
		else
		{
			fprintf(stderr,"  ???\n");
		}
	}
}

#if FALSE
#include <malloc.h>
static void my_init_hook(void);
static void *my_malloc_hook(size_t, const void *);
static void *(*old_malloc_hook)(size_t, const void *);
void (*__malloc_initialize_hook) (void) = my_init_hook;

static void my_init_hook(void)
{
	old_malloc_hook = __malloc_hook;
	__malloc_hook = my_malloc_hook;
}

static void* my_malloc_hook(size_t size, const void *caller)
{
	__malloc_hook = old_malloc_hook;

	logBacktrace(size);
	void* result = malloc(size);

	old_malloc_hook = __malloc_hook;

	__malloc_hook = my_malloc_hook;
	return result;
}
#endif

void* FE_CDECL operator new( size_t byteCount ) throw (std::bad_alloc)
{
//	logBacktrace(byteCount);
	return fe::allocate( byteCount );
}
void* FE_CDECL operator new[]( size_t byteCount ) throw (std::bad_alloc)
{
	return operator new( byteCount );
}

void FE_CDECL operator delete( void* pMemory ) throw ()
{
	fe::deallocate( pMemory );
}
void FE_CDECL operator delete[]( void* pMemory ) throw ()
{
	operator delete( pMemory );
}

